//************************************************************************
//   Copyright (C) 2004-2006 Maxim I. Guisov                            //
//                                                                      //
// Permission to use, copy, modify, distribute and sell this software   //
// and its documentation for any purpose is hereby granted without fee, //
// provided that the above copyright notice appear in all copies and    //
// that both that copyright notice and this permission notice appear    //
// in supporting documentation. Author makes no representations         //
// about the suitability of this software for any purpose.              //
// It is provided "as is" without express or implied warranty.          //
//                                                                      //
//************************************************************************
#include "StdAfx.h"
#include "Block.h"
#include "..\\Navigator\\BoardNavigator.h"
#include "..\\Utils\\IOLog.h"


Block::Block( int id, int color, const Point& start, BoardNavigator* bnav )
: StoneRegion( id, color, bnav ), chainID(-1), 
expert_estimate(1), predicted_status(1.0), in_question(true)
{
    ExtendBlock( start );
}


Block::~Block()
{
    for ( int ci = cerEntries.size(); ci-- > 0; )
        delete cerEntries[ci];
}


// PURPOSE: given the starting position, extend the block in all directions,
// update block index board, detect neibours, count perimeter, calculate local
// majority in the manhattan distance of 2, count all liberties of the block
void Block::ExtendBlock( const Point& start )
{
    MarkerBoard temp;

    // extend, fill neibours and count perimeter; 
    navi->blip.nc_set( start, id() );
    points.push_back( start );
    temp.nc_mark( start );

    int pi = 0;
    while ( pi < points.size() )
    {
        if ( navi->expert && navi->expert->nc_marked(points[pi]) )
            expert_estimate = 0;

        for ( int n = 0; n < 4; n++ )
        {
            Point adj( points[pi].adjacent(n) );
            if ( !adj.ok() )
            {
                perimeter_++;
                continue;
            }

            if ( navi->board.nc_get(adj) == color() && navi->blip.nc_get(adj) == -1 )
            {
                points.push_back(adj); 
                navi->blip.nc_set(adj, id());
            }

            else if ( navi->board.nc_get(adj) != color() && !temp.nc_marked(adj) )
            {
                neibours.push_back(adj);
                perimeter_++;

                if ( navi->board.nc_empty(adj) ) 
                    liberties[0].push_back(adj);
            }
            temp.nc_mark(adj);
        }
        pi++;
    }

    local_majority = 0;    
    for ( pi = neibours.size(); pi-- > 0; )
    {
        Point p( neibours[pi] );

        // take into account local majority in manhattan distance 1
        if ( navi->board.enemy(p, color()) ) 
            local_majority++;

        for ( int n = 0; n < 4; n++ )
        {
            Point adj( p.adjacent(n) );
            if ( !adj.ok() || navi->blip.nc_get(adj) == id() )
                continue;

            if ( navi->board.nc_get(adj) == color() )
            {
                if ( navi->board.nc_empty(p) )
                {
                    friends.push_back(adj);
                    connections.push_back(p);
                }
                /*
                else if ( navi->board.enemy(p, color()) )
                {
                    cutpoints.push_back(p);
                    cutfriends.push_back(adj);
                }
                */
            }

            if ( temp.nc_marked(adj) )
                continue;

            if ( navi->board.nc_empty(adj) && navi->board.nc_empty(p) )
                // count second order liberties
                liberties[1].push_back(adj);
            else if ( navi->board.nc_get(adj) == color() )
                local_majority--;
            else if ( navi->board.enemy(adj, color()) )
                local_majority++;

            temp.nc_mark(adj);
        }
    }

    friends.remove_duplicates();
    connections.remove_duplicates();

    // count third order liberties of the block
    for ( pi = liberties[1].size(); pi-- > 0; )
        for ( int n = 0; n < 4; n++ )
        {
            Point adj( liberties[1][pi].adjacent(n) );
            if ( adj.ok() && !temp.nc_marked(adj) && navi->board.nc_empty(adj) )
            {
                liberties[2].push_back(adj);
                temp.nc_mark(adj);
            }
        }

    AnalyseGeometry();
}


// PURPOSE: calculates bounding box and centre-of-mass
void Block::AnalyseGeometry()
{
    cm_x = 19; cm_y = 19;
    bbL = 19; bbR = 0; bbT = 19; bbB = 0;

    for ( int pi = points.size(); pi-- > 0; )
    {
        Point p( points[pi] );

        // check whether this point shifts centre-of-mass of the block
        if ( p.hdist() < cm_x ) cm_x = p.hdist();
        if ( p.vdist() < cm_y ) cm_y = p.vdist();

        // check bounding box boundaries
        if ( p.col() < bbL ) bbL = p.col();
        if ( p.col() > bbR ) bbR = p.col();
        if ( p.row() < bbT ) bbT = p.row();
        if ( p.row() > bbB ) bbB = p.row();
    }
}


void Block::preprocess()
{
    FindAdjacentBlocksAndCERs();
    CountSpecialLiberties();

    const CER* cer = navi->cer_at( points[0], opposite(color()) );
    opp_cer_size = cer->points.size();
    opp_cer_perimeter = cer->perimeter();
}

void Block::update_props()
{
    FindWeakestAndStrongest();
    ExamineAdjCERs();
}


int Block::CountInternalDame( StoneBoard* terri )
{
    if ( liberties[0].size() == 1 || connections.size() != 1 )
        return 0;

    int internal_dame = 0;
    PointArray int_liberties;

    if ( terri->get(connections[0]) == color() )
        int_liberties.push_back(connections[0]);

    if ( int_liberties.size() != 1 ) return 0;

    // now check that there are no other internal liberties, except one connection point
    int ilib_counter = 0, pi = liberties[0].size();
    while ( pi-- > 0 )
    {
        if ( terri->nc_get(liberties[0][pi]) == color() &&
             navi->board.nc_empty(liberties[0][pi]) )
            ilib_counter++;
    }

    if ( ilib_counter > 1 ) return 0;

    // todo: check where other connections lead...
    // lets go over all friends; if the block has at least two shared points
    // with any other block adjacent to the POINT, then return
    // otherwise... there might be big circle... otherwise...

    StoneBoard temp(*terri);
    while ( int_liberties.size() == 1 )
    {
        Point p( int_liberties[0] );
        int_liberties.clear();

        temp.set(p, eEmpty);
        internal_dame ++;

        for ( int n = 0; n < 4; n++ )
        {
            Point adj( p.adjacent(n) );
            if ( !adj.ok() ) continue;

            // get all friendly blocks adjacent to this dame and check their
            // internal liberties; we've already set this point to eEmpty in terri,
            // so rechecking starting block won't harm
            if ( navi->board.nc_get(adj) == color() && temp.nc_get(adj) == color() )
            {
                PointArray &libs = navi->nc_block_at(adj)->liberties[0];

                for ( pi = libs.size(); pi-- > 0; )
                {
                    if ( temp.nc_get( libs[pi] ) == color() )
                        int_liberties.push_back( libs[pi] );
                }
            }
        }
        int_liberties.remove_duplicates();
    }

    return internal_dame * ( color() == eBlack ? 1 : -1 );
}


// PURPOSE: reexamine adjacent eyespace, recurrent data for the neural network
void Block::ExamineAdjCERs()
{
    int ci = cerEntries.size();
    while ( ci-- > 0 )
        delete cerEntries[ci];

    cerEntries.clear();

    eye_num = 0;
    eye_sum_size = 0;
    eye_sum_perimeter = 0;
    eye_sum_split_points = 0;

    pacer_num = 0;
    pacer_average_power = 0;
    pacer_sum_acc_size = 0;
    pacer_sum_acc_perimeter = 0;
    pacer_sum_acc_split_points = 0;
    pacer_sum_int_size = 0;
    pacer_sum_int_perimeter = 0;
    pacer_sum_int_split_points = 0;

    for ( ci = adjCERs.size() - 1; ci >= 0; ci-- )
    {
        CER* cer = navi->get_cer(adjCERs[ci]); assert( cer );
        CER_Block_Entry* entry = new CER_Block_Entry( *cer, *this );

        if ( !cer->opponents && cer->eye_points.size() > 0 )  // fully accessible CER - eye
        {
            eye_num ++;
            eye_sum_size += cer->eye_points.size();
            eye_sum_perimeter += cer->eye_perimeter;
            eye_sum_split_points += cer->eye_split_points;
        } else // partially accessible CER
        {
            pacer_num ++;
            pacer_average_power += cer->average_power;
            pacer_sum_acc_size += entry->acc_points.size();
            pacer_sum_acc_perimeter += entry->acc_perimeter;
            pacer_sum_acc_split_points += entry->acc_split_points;
            pacer_sum_int_size += entry->int_points.size();
            pacer_sum_int_perimeter += entry->int_perimeter;
            pacer_sum_int_split_points += entry->int_split_points;
        }
        cerEntries.push_back(entry);
    }

    if ( pacer_num )
        pacer_average_power /= pacer_num;
}


// PURPOSE: perform static analysis of pure eyespace; if based on this analysis
// the block has two definite eyes, its in_question status changes to false,
// and in the network analysis its predicted_status will always be 1.0;
// also, if the block is in atari, it is highly possible, that it is dead;
// however it can be a killing hane, for example... need to think
void Block::AnalyseStatus()
{
    //// examine eyespace of the block
    //if ( in_question && eye_num >= 2 )
    //{
    //    predicted_status = 1.0;
    //    in_question = false;
    //}

    // todo: implement static block pre-analysis
    if ( liberties[0].size() == 1 )
    {
        predicted_status = 0.1;
        in_question = false;
    }

    pacer_num = 0;
    for ( int ci = adjCERs.size(); ci-- > 0; )
    {
        if ( navi->get_cer( adjCERs[ci] )->opponents > 0 ) pacer_num++;
    }

    if ( pacer_num == 0 )
        in_question = false;

    // if made decision is not equal those of the expert, cry out loud!
    if ( !in_question && navi->expert && 
         fabs(expert_estimate - predicted_status) > 0.5001 )
    {
        char buf[256];
        sprintf( buf, "static life & death analysis mistake: block %d\n", id() );
        LogWarning() << buf;
    }
}



// PURPOSE: count protected liberties of the block - those liberties, which
// put the adjacent opponent block(-s) into atari if he moves there
void Block::CountSpecialLiberties()
{
    protected_liberties = 0;
    autoatari_liberties = 0;

    // this block should not be in atari
    if ( liberties[0].size() == 1 ) return;

    // count protected liberties
    int pi = liberties[0].size();
    while ( pi-- > 0 )
    {
        Point p( liberties[0][pi] );
        PointArray opp_libs;
        int adj_opponents = 0;
        bool f_continue = false;

        for ( int n = 0; n < 4; n++ )
        {
            Point adj( p.adjacent(n) );
            if ( !adj.ok() ) continue;

            if ( navi->board.nc_empty(adj) )
            {
                opp_libs.push_back(adj);
                continue;
            }

            PointArray &libs = navi->nc_block_at(adj)->liberties[0];

            // adjacent friendly blocks should not be in the atari
            // and adjacent enemy blocks should not be too free :)
            if ( ( navi->board.nc_get(adj) == color() && libs.size() == 1 ) ||
                 ( navi->board.enemy(adj, color()) && libs.size() > 2 ) )
            {
                f_continue = true;
                break;
            }

            if ( navi->board.enemy(adj, color()) )
            {
                opp_libs.insert( opp_libs.end(), libs.begin(), libs.end() );
                adj_opponents++;
            }
        }

        if ( f_continue )
            continue;

        opp_libs.remove_duplicates();

        if ( (adj_opponents >= 1 && opp_libs.size() <= 2) ||
             (adj_opponents == 0 && opp_libs.size() <= 1) )
        {
            protected_liberties ++;
        }
    }

    if ( liberties[0].size() > 2 ) return;

    // count auto-atari liberties
    for ( pi = liberties[0].size(); pi-- > 0; )
    {
        Point p( liberties[0][pi] );
        PointArray friends_libs;
        bool f_continue = false;
        int adj_friends = 0;

        for ( int n = 0; n < 4; n++ )
        {
            Point adj( p.adjacent(n) );
            if ( !adj.ok() ) continue;

            if ( navi->board.nc_empty(adj) )
            {
                friends_libs.push_back(adj);
                continue;
            }

            //Block* adj_block = navi->block_at(adj); assert( adj_block );
            PointArray &libs = navi->nc_block_at(adj)->liberties[0];

            if ( navi->board.enemy(adj, color()) && libs.size() == 1 )
            {
                f_continue = true;
                break;
            }

            if ( navi->board.nc_get(adj) == color() )    // include this blocks liberties
            {
                friends_libs.insert( friends_libs.begin(), libs.begin(), libs.end() );
                adj_friends++;
            }
        }

        if ( f_continue )
            continue;

        friends_libs.remove_duplicates(); 
        // contains blocks liberties, second order liberties of current point
        // and liberties of other friendly blocks linked through this point

        if ( friends_libs.size() == 2 ) 
            autoatari_liberties++;
    }

    terr_direct_liberties =
    terr_all_friend_liberties =
    terr_all_opp_liberties = 0;

    // count own liberties in all adjacent spaces-in-question;
    // also count all friendly and opponent libertis in those spaces
    for ( int si = adjESpaces.size(); si-- > 0; )
    {
        ESpace* esp = navi->nc_get_espace( adjESpaces[si] );

        if ( !esp->in_question )
            continue;

        terr_direct_liberties += adjESpaceLiberties[si];
        for ( int bi = esp->adjBlocks.size(); bi-- > 0; )
        {
            Block* blk = navi->nc_get_block( esp->adjBlocks[bi] );

            if ( blk->id() == id() )
                continue;

            if ( blk->color() == color() )
                terr_all_friend_liberties += esp->adjBlocksLiberties[bi];
            else
                terr_all_opp_liberties += esp->adjBlocksLiberties[bi];
        }
    }
}


void Block::FindWeakestAndStrongest()
{
    opp_w1 = opp_w2 = -1; fri_s1 = -1;

    if ( adjOppBlocks.size() == 0 ) return;
    if ( adjOppBlocks.size() == 1 )
    {
        opp_w1 = adjOppBlocks[0];
        return;
    }

    for ( int iter = 0; iter < 2; iter++ )
    {
        double min_status = 2.0;
        int min_lib = 300;
        int min_index = 0;

        for ( int bi = adjOppBlocks.size() - 1; bi >= 0; bi-- )
        {
            if ( bi == opp_w1 ) continue;

            Block& block = *( navi->nc_get_block( adjOppBlocks[bi] ) );

            if ( min_status * 100 + min_lib > 
                 block.predicted_status * 100 + block.liberties[0].size() )
            {
                min_status = block.predicted_status;
                min_lib = block.liberties[0].size();
                min_index = bi;
            }
        }

        if ( opp_w1 == -1 )
            opp_w1 = min_index;
        else
            opp_w2 = min_index;
    }

    opp_w1 = adjOppBlocks[opp_w1];
    opp_w2 = adjOppBlocks[opp_w2];

    if ( friends.size() == 0 ) return;

    double max_status = 0.0;
    int max_lib = 0;
    int max_index = 0;

    for ( int pi = friends.size() - 1; pi >= 0; pi-- )
    {
        Block& block = *( navi->nc_block_at( friends[pi] ) );

        if ( max_status * 100 + max_lib < 
             block.predicted_status * 100 + block.liberties[0].size() )
        {
            max_status = block.predicted_status;
            max_lib = block.liberties[0].size();
            max_index = block.id();
        }
    }

    fri_s1 = max_index;
}


int Block::SharedLiberties( const Block& block ) const
{
    PointArray TBL( block.liberties[0] );
    TBL.insert( TBL.begin(), liberties[0].begin(), liberties[0].end() );
    TBL.remove_duplicates();

    return block.liberties[0].size() + liberties[0].size() - TBL.size();
}

