//************************************************************************
//   Copyright (C) 2004-2006 Maxim I. Guisov                            //
//                                                                      //
// Permission to use, copy, modify, distribute and sell this software   //
// and its documentation for any purpose is hereby granted without fee, //
// provided that the above copyright notice appear in all copies and    //
// that both that copyright notice and this permission notice appear    //
// in supporting documentation. Author makes no representations         //
// about the suitability of this software for any purpose.              //
// It is provided "as is" without express or implied warranty.          //
//                                                                      //
//************************************************************************
#include "StdAfx.h"
#include "CER.h"
#include "..\\Navigator\\BoardNavigator.h"


CER::CER( int id, int border, const Point& start, BoardNavigator* bnav )
: Region(id, border, bnav), opponents(0)
{
    MarkerBoard temp;

    if ( border == eBlack ) Extend( start, &navi->berip, &temp );
    else
        if ( border == eWhite ) Extend( start, &navi->werip, &temp );
}


// PURPOSE: extends color-enclosed region with flood fill, till it meets border color
// also counts opponent stones in the region and perimeter of the region
void CER::Extend( const Point& pos, IndexBoard* index, MarkerBoard* temp )
{
    if ( navi->board.enemy(pos, color()) )
        opponents++;

    points.push_back(pos);
    index->nc_set(pos, id());
    temp->nc_mark(pos);

    for ( int n = 0; n < 4; n++ )
    {
        Point adj( pos.adjacent(n) );
        if ( !adj.ok() )
        {
            perimeter_++; 
            continue;
        }

        if ( navi->board.nc_get(adj) != color() && index->nc_get(adj) == -1 )
        {
            Extend( adj, index, temp );
        }
        else if ( navi->board.nc_get(adj) == color() && !temp->nc_marked(adj) )
        {
            neibours.push_back(adj);
            perimeter_++;
            temp->nc_mark(adj);
        }
    }
}


// PURPOSE: if there is no opponent stones in the region, it is treated as pure eye region;
// this function divides region points into true and false eye points, then saves info
// about full eye points, counts its perimeter and split points
// NOTE: during incremental classification this function may be called several times;
// it uses information about predicted statuses of the diagonal adjacent blocks
void CER::ExamineEyeShape()
{
    if ( opponents ) return;

    eye_points.clear();
    eye_perimeter = 0;
    eye_split_points = 0;

    MarkerBoard temp;
    temp.nc_mark(points);

    PointArray partial;

    int pi = points.size();
    while ( pi-- > 0 )
    {
        Point p( points[pi] );
        int diag_index = 0;

        for ( int n = 0; n < 4; n++ )
        {
            Point diag( p.diagonal(n) );
            if ( !diag.ok() ) continue;

            // uncontrolled free point adds 1 to diag_index;
            if ( navi->board.nc_empty(diag) && !navi->power.control(diag, color()) )
                diag_index += 1;

            // point, occupied by safe opponent block adds 2 to diag_index;
            if ( navi->board.enemy(diag, color()) && navi->nc_block_at(diag)->IsSafe() )
                diag_index += 2;
        }

        // divide eye points on true and false
        if ( p.hdist() == 0 || p.vdist() == 0 )
        {
            // near the edge
            if ( diag_index ) continue;
        } 
        else
        {
            // in the centre
            if ( diag_index > 2 ) continue;
        }

        partial.push_back(p);
        temp.nc_clear(p);
    }

    // points, adjacent to false are only partial - we don't include them into analysis
    for ( pi = partial.size(); pi-- > 0; )
    {
        bool downgrade = false;
        for ( int n = 0; n < 4; n++ )
            if ( temp.marked( partial[pi].adjacent(n) ) )
            {
                downgrade = true;
                break;
            }
        if ( !downgrade )
            eye_points.push_back( partial[pi] );
    }

    eye_perimeter = Region::CountPerimeter( eye_points );
    eye_split_points = Region::CountSplitPoints( eye_points );
}


// PURPOSE: calculates average influence over the region, then makes the value "friendly"
void CER::CalcAveragePower()
{
    average_power = 0;
    for ( int pi = points.size(); pi-- > 0; )
        average_power += navi->power.nc_get( points[pi] );

    average_power /= points.size();

    // make value one-sided (for block in question BIQ)
    // remember, that black stones radiate positive power, while white - negative
    // but it doesn't matter to neural network, so we need positive power of every BIQ
    if ( color() == eWhite )
        average_power *= -1;
}

void CER::update_props()
{
    CalcAveragePower();
    ExamineEyeShape();
}


CER_Block_Entry::CER_Block_Entry( const CER& cer, const Block& block )
: acc_perimeter(0), acc_split_points(0), int_perimeter(0), int_split_points(0)
{
    if ( !cer.opponents )   // fully accessible CER for every border block
    {
        acc_points.assign( cer.eye_points.begin(), cer.eye_points.end() );
        acc_perimeter = cer.eye_perimeter;
        acc_split_points = cer.eye_split_points;
        return;
    }

    // otherwise we should divide the region into two parts: accessible
    // (from the given block) and internal (all other points)
    MarkerBoard temp;
    temp.nc_mark( cer.points );

    int pi = block.liberties[0].size();
    while ( pi-- > 0 )
    {
        Point p( block.liberties[0][pi] );
        if ( temp.nc_marked(p) ) // FOUND first-level accessible point
        {
            // get it's empty block from the storage and fill acc_points from it
            ESpace *espace = cer.navi->espace_at(p); assert( espace );
            acc_points.insert( acc_points.end(), espace->points.begin(), espace->points.end() );

            // clear all points, accessible from this liberty
            temp.nc_clear( espace->points );
        }
    }

    // now fill internal points...
    for ( pi = cer.points.size(); pi-- > 0; )
        if ( temp.nc_marked( cer.points[pi] ) )
            int_points.push_back( cer.points[pi] );

    // ...and count perimeters and split_points
    acc_perimeter = Region::CountPerimeter( acc_points );
    acc_split_points = Region::CountSplitPoints( acc_points );

    int_perimeter = Region::CountPerimeter( int_points );
    int_split_points = Region::CountSplitPoints( int_points );
}