//************************************************************************
//   Copyright (C) 2004-2006 Maxim I. Guisov                            //
//                                                                      //
// Permission to use, copy, modify, distribute and sell this software   //
// and its documentation for any purpose is hereby granted without fee, //
// provided that the above copyright notice appear in all copies and    //
// that both that copyright notice and this permission notice appear    //
// in supporting documentation. Author makes no representations         //
// about the suitability of this software for any purpose.              //
// It is provided "as is" without express or implied warranty.          //
//                                                                      //
//************************************************************************
#include "StdAfx.h"
#include "Chain.h"
#include "..\\Navigator\\BoardNavigator.h"


Chain::Chain( int id, Block* start, BoardNavigator* bnav )
 : StoneRegion(id, start->color(), bnav)
{
    Extend( start );
}


void Chain::Extend( Block* block )
{
    points.assign( block->points.begin(), block->points.end() );
    neibours.assign( block->neibours.begin(), block->neibours.end() );
    liberties.assign( block->liberties[0].begin(), block->liberties[0].end() );
    blocks.push_back( block->id() );
    block->chainID = id();

    PointArray friends( block->friends );
    navi->cip.nc_setvec( points, id() );
    
    perimeter_ = block->perimeter() - block->neibours.size();

    int pi = 0;
    while ( pi < friends.size() )
    {
        Point p( friends[pi] );

        if ( p.ok() && navi->cip.nc_get(p) == -1 )
        {
            Block &fb = *navi->nc_block_at(p); 
            
            points.insert( points.end(), fb.points.begin(), fb.points.end() );
            friends.insert( friends.end(), fb.friends.begin(), fb.friends.end() );
            neibours.insert( neibours.end(), fb.neibours.begin(), fb.neibours.end() );
            liberties.insert( liberties.end(), fb.liberties[0].begin(), fb.liberties[0].end() );
            blocks.push_back( fb.id() );
            fb.chainID = id();

            navi->cip.nc_setvec( fb.points, id() );
            perimeter_ += fb.perimeter() - fb.neibours.size();
        }
        pi++;
    }
    
    neibours.remove_duplicates();
    liberties.remove_duplicates();

#ifdef DEBUG_TESTER
    printf( "liberties of chain <%d> : ", id() );
    for ( int lib = liberties.size(); lib-- > 0; )
        printf( "<%d:%d,%d> ", (int)liberties[lib], liberties[lib].row(), liberties[lib].col() );
    printf("\n");
#endif

    perimeter_ += neibours.size();
}


void Chain::preprocess()
{
    FindAdjacentBlocksAndCERs();

    ////// add territory accessment
    PointArray tcheck( liberties );
    int pi = liberties.size();
    while ( pi-- > 0 )
    {
        Point p( liberties[pi] );
        if ( navi->zTerritory.nc_color(p) == color() )
            territory.push_back(p);
    }

    MarkerBoard temp;
    temp.nc_mark( territory );

    for ( pi = 0; pi < territory.size(); pi++ )
    {
        Point p( territory[pi] );
        for ( int n = 0; n < 4; n++ )
        {
            Point adj( p.adjacent(n) );
            if ( !adj.ok() )
                continue;

            if ( !temp.nc_marked(adj) && navi->board.nc_empty(adj) &&
                 navi->zTerritory.nc_color(adj) == color() )
            {
                territory.push_back(adj);
                temp.nc_mark(adj);
            }
        }
    }

    FindClosestOppBlocks();
    CountExternalOppLiberties();
    CountSplitLiberties();
}

void Chain::update_props()
{
    FindWeakestOppBlock();
    CalcAverageStatus();
    CheckEyespace();
}


void Chain::CalcAverageStatus()
{
    average_status = 0;
    for ( int bi = blocks.size(); bi-- > 0; )
    {
        Block& block = *( navi->nc_get_block( blocks[bi] ) );
        average_status += block.predicted_status * block.points.size();
    }
    average_status /= points.size();
}


double Chain::GetMinimalStatus()
{
    double ms = 1.0;
    for ( int bi = blocks.size(); bi-- > 0; )
    {
        Block& block = *( navi->nc_get_block( blocks[bi] ) );
        if ( ms > block.predicted_status ) 
            ms = block.predicted_status;
    }
    return ms;
}

void Chain::CheckEyespace()
{
    eye_num = 0;
    eye_sum_size = 0;
    eye_sum_perimeter = 0;
    eye_sum_split_points = 0;

    pacer5_num = pacer9_num = 0;
    pacer5_avepow = pacer9_avepow = 0.0;

    for ( int ci = adjCERs.size(); ci-- > 0; )
    {
        CER* cer = navi->get_cer( adjCERs[ci] ); assert( cer );
        if ( !cer->opponents && cer->eye_points.size() )  // fully accessible CER
        {
            eye_num++;
            eye_sum_size += cer->eye_points.size();
            eye_sum_perimeter += cer->eye_perimeter;
            eye_sum_split_points += cer->eye_split_points;
            continue;    
        }
        else if ( cer->points.size() <= 5 )
        {
            pacer5_num++;
            pacer5_avepow += cer->average_power;
        }
        else if ( cer->points.size() <= 9 )
        {
            pacer9_num++;
            pacer9_avepow += cer->average_power;
        }
    }

    if ( pacer5_num ) pacer5_avepow /= pacer5_num;
    if ( pacer9_num ) pacer9_avepow /= pacer9_num;
}


// NOTE: slow!!!
void Chain::CountExternalOppLiberties()
{
    MarkerBoard temp;
    extOppLiberties.clear();

    // to find all external opponent liberties:
    // 1. mark all partially accessible CERs adjacent to the chain blocks;
    int ci = adjCERs.size();
    while ( ci-- > 0 )
    {
        CER* cer = navi->get_cer( adjCERs[ci] ); assert( cer );
        if ( !cer->opponents )  // fully accessible CER
        {
            eye_num++;
            eye_sum_size += cer->eye_points.size();
            eye_sum_perimeter += cer->eye_perimeter;
            eye_sum_split_points += cer->eye_split_points;
            continue;    
        }

        temp.nc_mark( cer->points );
    }

    // 2. analyse all acc_points for these CERs from these blocks and unmark them;
    int bi = blocks.size();
    while ( bi-- > 0 )
    {
        Block* block = navi->nc_get_block( blocks[bi] ); assert( block );
        for ( ci = block->cerEntries.size(); ci-- > 0; )
        {
            // skip fully accessible CER
            if ( !block->cerEntries[ci]->int_points.size() ) 
                continue;

            temp.nc_clear( block->cerEntries[ci]->acc_points );
        }
    }

    // 3. count marked liberties of the adjacent opponents blocks;
    for ( bi = adjOppBlocks.size(); bi-- > 0; )
    {
        PointArray &libs = navi->nc_get_block( adjOppBlocks[bi] )->liberties[0];
        for ( ci = libs.size(); ci-- > 0; )
        {
            Point p( libs[ci] );
            if ( temp.nc_marked(p) ) 
                extOppLiberties.push_back(p);
        }
    }
}


void Chain::FindClosestOppBlocks()
{
    MarkerBoard temp;

    // mark all adjacent opponent blocks
    int bi = adjOppBlocks.size();
    while ( bi-- > 0 )
        temp.nc_mark( navi->nc_get_block( adjOppBlocks[bi] )->points );

    // for each block in chain
    for ( bi = blocks.size(); bi-- > 0; )
    {
        PointArray &libs = navi->nc_get_block( blocks[bi] )->liberties[0];
        for ( int pi = libs.size(); pi-- > 0; )
        {
            Point p( libs[pi] );

            // for each adjacent point of liberty
            for ( int n = 0; n < 4; n++ )
            {
                Point adj( p.adjacent(n) );
                if ( adj.ok() && !temp.nc_marked(adj) && !navi->board.nc_empty(adj) )
                {
                    const Block* blk = navi->nc_block_at( adj );
                    temp.nc_mark( blk->points );
                    shaOppBlocks.push_back( blk->id() );
                }
            }
        }
    }
}


void Chain::FindWeakestOppBlock()
{
    opp_w3 = -1;
    double min_status = 2.0;
    int min_lib = 300;
    int min_index = 0;
    
    // check adjacent blocks first
    if ( adjOppBlocks.size() == 1 ) opp_w3 = adjOppBlocks[0]; else
    if ( adjOppBlocks.size() >= 2 )
    {
        for ( int bi = adjOppBlocks.size(); bi-- > 0; )
        {
            Block& block = *( navi->nc_get_block( adjOppBlocks[bi] ) );
            if ( min_status >  block.predicted_status || 
               ( min_status == block.predicted_status &&
                 min_lib > block.liberties[0].size() ))
            {
                min_status = block.predicted_status;
                min_lib = block.liberties[0].size();
                min_index = bi;
            }
        }
        opp_w3 = adjOppBlocks[min_index];
    }


    double min_status2 = 2.0;
    int min_lib2 = 300;
    if ( shaOppBlocks.size() == 0 ) return;

    for ( int bi = shaOppBlocks.size(); bi-- > 0; )
    {
        Block& block = *( navi->nc_get_block( shaOppBlocks[bi] ) );
        if ( min_status2 >  block.predicted_status || 
           ( min_status2 == block.predicted_status &&
             min_lib2 > block.liberties[0].size() ))
        {
            min_status2 = block.predicted_status;
            min_lib2 = block.liberties[0].size();
            min_index = bi;
        }
    }
    
    if ( opp_w3 == -1 || min_status2 * 100 + min_lib2 < min_status * 100 + min_lib )
        opp_w3 = shaOppBlocks[min_index];
}


void Chain::CountSplitLiberties()
{
    PointArray allpts(points);
    allpts.insert( allpts.begin(), liberties.begin(), liberties.end() );

    PointArray allsplits;
    Region::CountSplitPoints( allpts, &allsplits );

    MarkerBoard temp;
    temp.nc_mark( allsplits );

    for ( int pi = liberties.size(); pi-- > 0; )
        if ( temp.nc_marked( liberties[pi] ) )
            split_liberties.push_back(pi);
}

