//************************************************************************
//   Copyright (C) 2004-2006 Maxim I. Guisov                            //
//                                                                      //
// Permission to use, copy, modify, distribute and sell this software   //
// and its documentation for any purpose is hereby granted without fee, //
// provided that the above copyright notice appear in all copies and    //
// that both that copyright notice and this permission notice appear    //
// in supporting documentation. Author makes no representations         //
// about the suitability of this software for any purpose.              //
// It is provided "as is" without express or implied warranty.          //
//                                                                      //
//************************************************************************
#ifndef POINT_H_INCLUDED_
#define POINT_H_INCLUDED_

#include <vector>

const int cBoardSize = 361; // 19*19;

typedef unsigned long ULONG;
typedef signed char SBYTE;

enum { bSize1D = 19, bSize2D = 361 };

// BIT LOGIC
const ULONG cPointBit[bSize1D] =
{
    0x00001, 0x00002, 0x00004, 0x00008, 
    0x00010, 0x00020, 0x00040, 0x00080,
    0x00100, 0x00200, 0x00400, 0x00800,
    0x01000, 0x02000, 0x04000, 0x08000,
    0x10000, 0x20000, 0x40000
};

const ULONG cPointMask = 0x4ffff;


enum tColor { eEmpty = 0, eBlack = 1, eWhite = 2 };

inline int opposite( int color ) {
    return (color == eBlack || color == eWhite) ? (3 - color) : -1;
}


// this class represents a point on the 19x19 (hardcoded) board
class Point
{
public:

    Point( int pos = 0 ) : pos_(pos) {}
    Point( const Point& p ) : pos_(p.pos_) {}
    Point( int col, int row ) : pos_(row*19 + col) {}

    //bool operator == ( const Point &p ) { return pos_ == p.pos_; }

    // checks whether the point is on board
    bool ok() const { return ( pos_ >= 0 && pos_ < cBoardSize ); }

    // supposed to return column of the point
    int col() const { return pos_ % 19; }

    // supposed to return row of the point
    int row() const { return pos_ / 19; }

    // minimal distance to the left/right edge
    int hdist() const;

    // minimal distance to the top/bottom edge
    int vdist() const;
    
    // returns adjacent points: 0 - left, 1 - top, 2 - right, 3 - bottom
    const Point& adjacent( int n ) const
    { return Point::adjacent_[pos_][n]; }

    // returns diagonal adjacent points: 0 - TL, 1 - TR, 2 - BR, 3 - BL
    const Point& diagonal( int n ) const
    { return Point::diagonal_[pos_][n]; }

    operator int () const { return pos_; }

private:

    friend class PointShiftCache;

    // adjacent points accessors
    const Point L() const { if (pos_ % 19 ==  0) return -1; return pos_ - 1; }
    const Point R() const { if (pos_ % 19 == 18) return -1; return pos_ + 1; }
    const Point T() const { return pos_ - 19; }
    const Point B() const { return pos_ + 19; }

    int pos_;

    static PointShiftCache cache_;
    static Point adjacent_[cBoardSize][4];
    static Point diagonal_[cBoardSize][4];
};


class PointShiftCache
{
public:
    // fills adjacent_ and diagonal_ caches
    PointShiftCache();

private:

    static Point adjacent( Point pos, int n );
    static Point diagonal( Point pos, int n );
};

// simple and plain point storage
//typedef std::vector<Point> PointArray;
class PointArray: public std::vector<Point>
{
public:

    void remove_duplicates();
    static bool greater( const Point& p1, const Point& p2 );
    static bool equal( const Point& p1, const Point& p2 );
};


// widely used index storage: objects remember index of other object in the storage;
// this way it is harder to adress the object, but much easier to delete
typedef std::vector<int> IndexArray;


#endif // POINT_H_INCLUDED_