//************************************************************************
//   Copyright (C) 2004-2006 Maxim I. Guisov                            //
//                                                                      //
// Permission to use, copy, modify, distribute and sell this software   //
// and its documentation for any purpose is hereby granted without fee, //
// provided that the above copyright notice appear in all copies and    //
// that both that copyright notice and this permission notice appear    //
// in supporting documentation. Author makes no representations         //
// about the suitability of this software for any purpose.              //
// It is provided "as is" without express or implied warranty.          //
//                                                                      //
//************************************************************************
#include "StdAfx.h"
#include "BoardNavigator.h"
#include "BlockFeatures.h"
#include "..\\RPropNet\\RPropNet.h"
#include "..\\Utils\\IOLog.h"
#include "..\\Utils\\NetConfig.h"
#include "..\\Utils\\MorphEngine.h"

#pragma warning ( disable : 4244 )

BoardNavigator::BoardNavigator( const StoneBoard& EndPosition, const NetFactory* nf, int brdId )
: netFactory_(nf), board(EndPosition), expert(NULL), blip(-1), berip(-1), werip(-1), cip(-1)
{
    isTestingBoard_ = false;
    boardId = brdId;
}

BoardNavigator::~BoardNavigator()
{
    int i;
    for ( i = allRegions.size(); i-- > 0; ) delete allRegions[i];
    for ( i = allChains.size(); i-- > 0; )  delete allChains[i];
    for ( i = allCERs.size(); i-- > 0; )    delete allCERs[i];
}


int BoardNavigator::ScoreEndOfGame( const StoneBoard& EndPosition, const NetFactory* nf )
{
    //BoardNavigator BNav( EndPosition, nf );
    //BNav.Preprocess();
    //BNav.ProcessBoard();
    //return BNav.CountTerritory();
    return 0;
}


// PURPOSE: divides the board into regions: blocks, chains, empty spaces and CERs
// then analyses common static features of the blocks and chains
void BoardNavigator::Preprocess()
{
    MorphEngine::EstimateInfluence( &zInfluence, board );
    MorphEngine::EstimateTerritory( &zTerritory, board );

    // divide the board into blocks and CERs
    for ( int pos = 0; pos < cBoardSize; pos++ )
    {
        if ( blip.nc_get(pos) == -1 )
        {
            if ( board.empty(pos) )
            {
                allRegions.push_back( new ESpace( allRegions.size(), pos, this ));
                esIndex.push_back( allRegions.size() - 1 );
            }
            else
            {
                allRegions.push_back( new Block( allRegions.size(), board.nc_get(pos), pos, this ));
                sbIndex.push_back( allRegions.size() - 1 );
            }
        }

        if ( !board.black(pos) && berip.nc_get(pos) == -1 )
            allCERs.push_back( new CER( allCERs.size(), eBlack, pos, this ));

        if ( !board.white(pos) && werip.nc_get(pos) == -1 )
            allCERs.push_back( new CER( allCERs.size(), eWhite, pos, this ));
    }

    // concatenate all blocks with shared liberties into chains
    for ( unsigned int bi = 0, bimax = sbIndex.size(); bi < bimax; bi++ )
    {
        Block* block = nc_get_block( sbIndex[bi] ); assert( block );
        if ( block->chainID == -1 )
            allChains.push_back( new Chain( allChains.size(), block, this ));
    }

    // analyse questionable territory, select BIQs
    CheckTerritoryAndSelectBIQs();
}


// PURPOSE: examines all questionable territories, and fills BIQs index array
void BoardNavigator::CheckTerritoryAndSelectBIQs()
{
    unsigned int bi;

    // examine adjacent blocks, count their liberties in the ESpace
    for ( bi = esIndex.size(); bi-- > 0; )
        nc_get_espace( esIndex[bi] )->preprocess();

    // examine adjacent points, count protected liberties
    for ( bi = sbIndex.size(); bi-- > 0; )
        nc_get_block( sbIndex[bi] )->preprocess();

    // examine adjacent blocks, CERSs, count external opponents liberties
    for ( bi = allChains.size(); bi-- > 0; )
        allChains[bi]->preprocess();

    // let's consider all blocks as BIQ if they cant argue
    for ( bi = sbIndex.size(); bi-- > 0; )
    {
        Block* block = nc_get_block( sbIndex[bi] );

        // let block try and stand for itself: perform static analysis of its L&D
        block->AnalyseStatus(); 

        // if block is still in question, put it under neural classifier
        if ( block->in_question )
            biqIndex.push_back( block->id() );
    }
}

void BoardNavigator::ProcessBoard( const NetConfig &netConfig )
{
	int nnum = netFactory_->NetsNumber();
    for ( int i = 0; i < nnum; i++ )
    {
        netError = 0;
        PerformGradingStep( i, netConfig );
    }
}

// PURPOSE: update dynamical block-in-question props
void BoardNavigator::PrepareGradingStep()
{
    // calculate power of the stones over the board
    CalcRoughInfluence();

    // calculate dynamic numerical features of the CERs, blocks and chains
    int bi;
    for ( bi = allCERs.size(); bi-- > 0; )
        allCERs[bi]->update_props();

    for ( bi = allChains.size(); bi-- > 0; )
        allChains[bi]->update_props();

    for ( bi = sbIndex.size(); bi-- > 0; )
        nc_get_block( sbIndex[bi] )->update_props();
}

// PURPOSE: performs one step of recurrent analysis with neural net
void BoardNavigator::PerformGradingStep( int iteration, const NetConfig &netConfig )
{
    PrepareGradingStep();

#ifdef DEBUG_TESTER
    int aliveCounter = 0, deadCounter = 0;
#endif

    int errBlockEvalCounter = 0;
    for ( int bi = biqIndex.size(); bi-- > 0; )
    {
        Block& block = *nc_get_block( biqIndex[bi] );

        static BFeatures bf( netConfig );
        bf.fill( block, iteration );

        double io_status = 0;

        if ( expert ) 
        {
            io_status = block.expert_estimate;
            netFactory_->Network(iteration)->apply( bf.fvector, &io_status, &io_status, isTestingBoard_ );
            block.predicted_status = io_status;
        }
        else
            netFactory_->Network(iteration)->apply( bf.fvector, &io_status, 0, true );

#ifdef DEBUG_TESTER
        printf( "block <%d>: status = %g\n\n", block.id(), io_status );

        if ( io_status > 0.5 )
            aliveCounter++;
        else
            deadCounter++;
#endif

        // use 40-20-40 error evaluation scheme
        if ( expert )
        {
            if ( ( block.predicted_status < 0.6 && block.expert_estimate == 1 ) ||
                 ( block.predicted_status > 0.4 && block.expert_estimate == 0 ) )
                errBlockEvalCounter++;
        }
    }

#ifdef DEBUG_TESTER
    if ( expert )
    {
		printf( "%d blocks in all, %d BIQs, %d predicted ALIVE, %d predicted DEAD, %d incorrect",
                sbIndex.size(), AliveCounter + DeadCounter, AliveCounter, DeadCounter, NetworkErrorCounter );
    }
#endif

    netError += errBlockEvalCounter;
}



int BoardNavigator::CountTerritory()
{
    StoneBoard final( board );
    int territory_count = 0;
    int dead_black = 0;
    int dead_white = 0;

#ifdef DEBUG_TESTER
    print();
#endif

    // remove dead chains from the board.
    for ( unsigned int ci = 0; ci < allChains.size(); ci++ )
    {
        Chain* chain = allChains[ci];
        if ( final.empty( chain->points[0] ) ) continue;
        chain->CalcAverageStatus();

        if ( chain->average_status < 0.5 )
        {
            if ( chain->eye_num >= 2 )
                LogWarning() << "a group with at least two eyes was considered dead";

            if ( chain->color() == eBlack ) dead_black += chain->points.size();
            if ( chain->color() == eWhite ) dead_white += chain->points.size();
            final.setvec( chain->points, eEmpty );

#ifdef DEBUG_TESTER
            printf( "Dead at (%d,%d), size = %d\n",
                    chain->points[0].col()+1, chain->points[0].row()+1, chain->points.size() );
#endif
        }
    }

    // count territories
    MarkerBoard temp;
    StoneBoard terri(final);
    for ( int pos = 0; pos < cBoardSize; pos++ )
    {
        if ( final.empty(pos) && !temp.nc_marked(pos) )
        {

            PointArray points;
            points.push_back(pos);
            temp.nc_mark(pos);
            int adj_blacks = 0, adj_whites = 0;

            unsigned int pi = 0;
            while ( pi < points.size() )
            {
                for ( unsigned int n = 0; n < 4; n++ )
                {
                    Point adj( points[pi].adjacent(n) );

                    if ( !adj.ok() ) continue;

                    if ( final.empty(adj) && !temp.nc_marked(adj) )
                    {
                        points.push_back(adj);
                        temp.nc_mark(adj);
                    }

                    else if ( final.black(adj) )
                    {
                        adj_blacks ++; 
                    }

                    else if ( final.white(adj) )
                    {
                        adj_whites ++;
                    }
                }
                pi++;
            }

            if ( (adj_blacks > 0) && (adj_whites == 0) )
            {
                territory_count += points.size(); 
                terri.setvec( points, eBlack );
            }
            else if ( (adj_whites > 0) && (adj_blacks == 0) )
            {
                territory_count -= points.size();
                terri.setvec( points, eWhite );
            }
        }
    }


    // check those block's liberties, that has only connection value, and not a point
    for ( unsigned int bi = 0; bi < sbIndex.size(); bi++ )
    {
        Block* block = get_block( sbIndex[bi] ); assert( block );

        if ( !final.empty( block->points[0] ) )  // alive block
        {
            int dame_connections = block->CountInternalDame( &terri );
            territory_count -= dame_connections;

#ifdef DEBUG_TESTER
            if ( dame_connections ) printf( "Dame connections: %d, for block at (%d,%d)\n",
                                            dame_connections, block->points[0].col()+1, block->points[0].row()+1 );
#endif
        }
    }


#ifdef DEBUG_TESTER
    for ( int r = 0; r < 19; r++ )
    {
        for ( int c = 0; c < 19; c++ )
        {
            Point p (c, r);

            char ch = '.';
            if ( final.black(p) ) ch = '*';
            else
                if ( final.white(p) ) ch = 'o';
            else
                if ( final.empty(p) )
            {
                if ( terri.black(p) ) ch = '+';
                else
                    if ( terri.white(p) ) ch = '-';
            }

            printf( " %c", ch );
        }
        printf( "\n" );
    }

    printf( "ON BOARD: dead B(%d), dead W(%d), territory balance(%d)\n\n",
            dead_black, dead_white, territory_count );

#endif

    return territory_count + dead_white - dead_black;
}


// PURPOSE: calculates influence of the stones over the board
// one stone radiates it's power over manhattan distance of 5
void BoardNavigator::CalcRoughInfluence()
{
    power.clear();
    
    // variant 1: use zobrist influence function
    // MorphEngine::EstimateInfluence( &power, board );

    // variant 2: calc influence manually
    for ( int pos = 0; pos < cBoardSize; pos++ )
    {
        Point p(pos);
        if ( board.empty(p) ) continue;

        // black positive, white negative
        double z = ( board.black(p) ? 1.0 : -1.0 ) * nc_block_at(p)->predicted_status;

        for ( int n = 0; n < 4; n++ )
        {
            Point p_a = p.adjacent(n);
            if ( p_a.ok() )
            {
                power.nc_inc( p_a, 64*z );
                Point p_aa = p_a.adjacent(n);
                if ( p_aa.ok() )
                {
                    power.nc_inc( p_aa, 32*z );
                    Point p_aaa = p_aa.adjacent(n);
                    if ( p_aaa.ok() )
                    {
                        power.nc_inc( p_aaa, 20*z );
                        Point p_aaaa = p_aaa.adjacent(n);
                        if ( p_aaaa.ok() )
                        {
                            power.nc_inc( p_aaaa, 14*z );
                            power.inc( p_aaaa.adjacent(n), 11*z );
                        }

                        Point p_aaab = p_aaa.adjacent(n+1);
                        if ( p_aaab.ok() )
                        {
                            power.nc_inc( p_aaab, 16*z );
                            power.inc( p_aaab.adjacent(n), 12*z );
                        }
                    }

                    Point p_aab  = p_aa.adjacent(n+1);
                    if ( p_aab.ok() )
                    {
                        power.nc_inc( p_aab, 26*z );
                        Point p_aabb = p_aab.adjacent(n+1);
                        if ( p_aabb.ok() )
                        {
                            power.nc_inc( p_aabb, 16*z );
                            power.inc( p_aabb.adjacent(n), 12*z );
                        }
                    }
                }

                Point p_ab = p_a.adjacent(n+1);
                if ( p_ab.ok() )
                {
                    power.nc_inc( p_ab, 40*z );
                    Point p_abb = p_ab.adjacent(n+1);
                    if ( p_abb.ok() )
                    {
                        power.nc_inc( p_abb, 26*z );
                        Point p_abbb = p_abb.adjacent(n+1);
                        if ( p_abbb.ok() )
                        {
                            power.nc_inc( p_abbb, 16*z );
                            power.inc( p_abbb.adjacent(n), 12*z );
                            power.inc( p_abbb.adjacent(n+1), 12*z );
                        }
                    }
                }
            }
        } // for (n)
    } // for (pos)
}


// PURPOSE: pure debug function, used to display different planes
void BoardNavigator::print()
{
    printf("\n\n\n");

    for ( int r = 0; r < 19; r++ )
    {
        for ( int c = 0; c < 19; c++ )
        {
   //         Point pos(c,r);
   //         char ch;

			//// influence
   //         ch = ( power.control(pos, eBlack) ? '+' :
   //              ( power.control(pos, eWhite) ? '-' : ' ' ));
   //         if ( expert && expert->get(pos) == eWhite ) ch = 'o';
   //         if ( expert && expert->get(pos) == eBlack ) ch = '*';
   //         
   //         // influence
   //         ch = ( power.control(pos, eBlack) ? '+' :
   //              ( power.control(pos, eWhite) ? '-' : ' ' ));
   //         if ( board.white(pos) ) ch = 'o';
   //         if ( board.black(pos) ) ch = '*';

			//// blocks
   //         ch = 'a';
   //         ch += blip.get(pos);

			//// black CERs
   //         ch = 'a';
   //         ch += werip.nc_get(pos);

			//// white CERs
   //         ch = '#'+1;
   //         ch += werIndex.get(pos);

			//// chains
   //         ch = 'a';
   //         ch += cip.get(pos);

			////printf(" %c", ch);
   //         printf( "%3d", blip.get(pos) );
        }
        printf("\n");
    }
}



// different objects accessors
ESpace* BoardNavigator::espace_at( const Point& pos )
{
    if ( !pos.ok() || !board.empty(pos) ) return NULL;
    return static_cast<ESpace*> (allRegions[blip.get(pos)]);
}

Block* BoardNavigator::block_at( const Point& pos )
{
    if ( !pos.ok() || board.empty(pos) ) return NULL;
    return static_cast<Block*> (allRegions[blip.get(pos)]);;
}

Chain* BoardNavigator::chain_at( const Point& pos )
{
    return pos.ok() && !board.empty(pos) ? allChains[cip.get(pos)] : NULL; 
}

CER* BoardNavigator::cer_at( const Point& pos, int color )
{
    if ( !pos.ok() || color == eEmpty ) return NULL;
    return color == eBlack ? allCERs[berip.get(pos)] : allCERs[werip.get(pos)];
}

ESpace* BoardNavigator::get_espace( unsigned int index )
{
    if ( index >= 0 && index < allRegions.size() && allRegions[index]->color() == eEmpty )
        return static_cast<ESpace*> (allRegions[index]);
    return NULL;
}

Block* BoardNavigator::get_block( unsigned int index )
{
    if ( index >= 0 && index < allRegions.size() && allRegions[index]->color() != eEmpty )
        return static_cast<Block*> (allRegions[index]);
    return NULL;
}

Chain* BoardNavigator::get_chain( unsigned int index )
{
    if ( index >= 0 && index < allChains.size() )
        return allChains[index];
    return NULL;
}

CER* BoardNavigator::get_cer( unsigned int index )
{
    if ( index >= 0 && index < allCERs.size() )
        return allCERs[index];
    return NULL;
}


void TrainingFactory::AddFinalPosition( const StoneBoard &EndPosition, const MarkerBoard &ExpertBoard )
{
    positions_.push_back( new StoneBoard( EndPosition ) );
    experts_.push_back( new MarkerBoard( ExpertBoard ) );
    static double err0 = -1;
    if ( netConfig_->getStartOnHalfTS() )
        err0 = -1 - err0;
    errors_.push_back( err0 );
}


void TrainingFactory::ResetErrors()
{
    static int ne = 0;
    for ( int i = errors_.size(); i-- > 0; )
        if ( netConfig_->getStartOnHalfTS() )
        {
            errors_[i] = ne;
            ne = -1 - ne;   // 0 => -1 => 0 => -1 and so on..
        }
        else
            errors_[i] = -1;
}

void TrainingFactory::Process()
{
    printf("processing [");
    double p = 0;
    int pi = 0;

    //sumError_ = 0;
    int imax = positions_.size();
    for ( int i = 0; i < imax; i++ )
    {
        // each 10th board belongs to testing set
        bool isTB = ( i % 10 == 0 );
        
        // don't skip testing boards evaluation
        if ( errors_[i] != 0 || !netConfig_->getDropSolved() || isTB )
        {
            BoardNavigator nav( *positions_[i], netFactory_ );
            nav.expert = experts_[i];
            nav.isTestingBoard_ = isTB;
            nav.Preprocess();
            nav.ProcessBoard( *netConfig_ );
            errors_[i] = nav.netError;
            //sumError_ += errors_[i];
        }
        else if ( netConfig_->getRefreshDrop() )
            errors_[i] = -1;

        if ( ( p += 50.0 / imax ) > pi )
        {
            printf(".");
            pi++;
        }
    }
    printf("]\n");
}

TrainingFactory::~TrainingFactory()
{
    for ( int i = positions_.size() - 1; i >= 0; i-- )
    {
        delete static_cast<StoneBoard*>( positions_[i] );
        delete static_cast<MarkerBoard*>( experts_[i] );
    }

    positions_.clear();
    experts_.clear();
    
    for ( int i = features_.size() - 1; i >= 0; i-- )
        delete static_cast<BFeatures*>( features_[i] );

    features_.clear();
}


void BoardNavigator::PreprocessAndStoreFeatures( std::vector< BFeatures* > &features, const NetConfig &netConfig )
{
    Preprocess();
    
    // cascade networks only train last one
    int n = 0;
    for ( ; n < netFactory_->NetsNumber() - 1; n++ )
    {
        PrepareGradingStep();
        for ( int bi = biqIndex.size(); bi-- > 0; )
        {
            Block &block = *nc_get_block( biqIndex[bi] );

            BFeatures bf( netConfig );
            bf.fill( block, n );
            
            double est = block.expert_estimate;
            netFactory_->Network(n)->apply( bf.fvector, &block.predicted_status, &est, true );
        }        
    }

    PrepareGradingStep();
    for ( int bi = biqIndex.size(); bi-- > 0; )
    {
        Block& block = *nc_get_block( biqIndex[bi] );

        features.push_back( new BFeatures( netConfig ) );
        features.back()->fill( block, n );
    }        
}


void TrainingFactory::ParsePositionAndStoreFeatures( 
    const StoneBoard &EndPosition, 
    const MarkerBoard &ExpertBoard,
    int boardId )
{
    BoardNavigator nav( EndPosition, netFactory_, boardId );
    nav.expert = &ExpertBoard;
    nav.PreprocessAndStoreFeatures( features_, *netConfig_ );
}


void TrainingFactory::ProcessParsed( bool testIt )
{
    if ( !testIt )
        printf("processing [");
    double p = 0;
    int pi = 0;

    int imax = features_.size();
    int blockTrnErrorCounter = 0;
    int blockTstErrorCounter = 0;
    for ( int i = 0; i < imax; i++ )
    {
        // each 10th block belongs to testing set
        bool isTB = ( i % 10 == 0 );
        
        BFeatures &bf = *features_[i];
        double ret;
        netFactory_->Last()->apply( bf.fvector, &ret, &bf.expert_estimate, isTB );
        
        if ( ( ret - 0.5 ) * ( bf.expert_estimate - 0.5 ) < 0 )
        //if ( ( ret < 0.6 && bf.expert_estimate > 0.6 ) ||
        //     ( ret > 0.4 && bf.expert_estimate < 0.4 ) )
        {
            if ( isTB )
                blockTstErrorCounter++;
            else
                blockTrnErrorCounter++;
                
            if ( testIt )
                printf( "%s board <%d> block at <%d, %d>: est = %2.2f res = %2.2f\n", 
                    ( isTB ? "tst" : "trn" ), bf.boardId, bf.x, bf.y, bf.expert_estimate, ret );
        }
        
        if ( !testIt && ( p += 50.0 / imax ) > pi )
        {
            printf(".");
            pi++;
        }
    }
    if ( !testIt )
        printf("] %d + %d / %d\n", blockTrnErrorCounter, blockTstErrorCounter, imax);
    
    FILE *file = 0;
    std::ostringstream fname;
    fname << netFactory_->Last()->filename_ << ".out";
    
    fopen_s( &file, fname.str().c_str(), "at" );
    if ( !file )
    {
        printf("!!\n");
        return;
    }
    
    //fprintf( file, "%5.3f, %5.3f\n", 
    //    (float) blockTrnErrorCounter * 100.0 / (float) imax,
    //    (float) blockTstErrorCounter * 100.0 / (float) ( imax / 10 ) );
    
    fprintf( file, "%7.5f, %7.5f\n", 
        netFactory_->Last()->netTrainError() * 100.0, 
        netFactory_->Last()->netTestError() * 100.0 );

    fclose( file );    
}


