//************************************************************************
//   Copyright (C) 2004-2006 Maxim I. Guisov                            //
//                                                                      //
// Permission to use, copy, modify, distribute and sell this software   //
// and its documentation for any purpose is hereby granted without fee, //
// provided that the above copyright notice appear in all copies and    //
// that both that copyright notice and this permission notice appear    //
// in supporting documentation. Author makes no representations         //
// about the suitability of this software for any purpose.              //
// It is provided "as is" without express or implied warranty.          //
//                                                                      //
//************************************************************************
#ifndef BOARDS_H_INCLUDED_
#define BOARDS_H_INCLUDED_

#include "..\\Elements\\Point.h"
#pragma warning ( disable : 4275 )

template<class T> class IBoard
{
protected:
    T data_[ cBoardSize ];

    virtual char get_symbol( int p, void* refData = NULL ) const
    { return '.'; }

public:

    // filling ctr
    IBoard( T initval = 0 )
    {
        if ( initval == 0 ) memset( data_, 0, sizeof(T) * cBoardSize ); 
        else for ( int p = 0; p < cBoardSize; p++ ) data_[p] = initval; 
    }

    // copy ctr
    IBoard( const IBoard<T>& board )
    {
        copy( board );
    }

    void copy( const IBoard<T>& board )
    { 
        memcpy( data_, board.data_, sizeof(T) * cBoardSize );
    }

    T get( const Point& p ) const
    { return p.ok() ? data_[p] : -1; }

    T nc_get( int p ) const
    { return data_[p]; }

    void set( const Point& p, T val )
    { if ( p.ok() ) data_[p] = val; }

    void nc_set( int p, T val )
    { data_[p] = val; }

    void setvec( const PointArray& points, T val )
    { for ( int pi = points.size(); pi > 0; ) set( points[--pi], val ); }

    void nc_setvec( const PointArray& points, T val )
    { for ( int pi = points.size(); pi > 0; ) nc_set( points[--pi], val ); }

    void print( void* refData = NULL )
    {
        for ( int r = 0, p = 0; r < 19; printf("\n"), r++ )
            for ( int c = 0; c < 19 ; printf(" %c", get_symbol(p++, refData)), c++ ) ;
    }
};


// first-level indexing service for differend kinds of board features
class IndexBoard: public IBoard<int>
{
public:
    IndexBoard( int initval = 0 ) : IBoard<int>( initval ) {}
};


// advanced board, designed to ease stone solor assessment - Stone Plane
class StoneBoard: public IBoard<SBYTE>
{
protected:
    // defined in point.c
    virtual char get_symbol( int p, void* refData = NULL ) const;

public:

    StoneBoard( SBYTE initval = 0 ) : IBoard<SBYTE>( initval ) {}

    bool empty( const Point& p ) const { return get(p) == eEmpty; }
    bool black( const Point& p ) const { return get(p) == eBlack; }
    bool white( const Point& p ) const { return get(p) == eWhite; }

    bool nc_empty( const Point& p ) const { return nc_get(p) == eEmpty; }
    bool nc_black( const Point& p ) const { return nc_get(p) == eBlack; }
    bool nc_white( const Point& p ) const { return nc_get(p) == eWhite; }

    bool enemy( const Point& p, int color ) const
    { return !p.ok() ? false : nc_get(p) == opposite(color); }
};


// masked variation of the stoneboard helps with some morphing algorithms
class StoneBoardMasked: public StoneBoard
{
protected:
    
    void mask_set( const Point& p, SBYTE val ) 
    {
        if ( val == eEmpty ) mask[p.row()] &= !cPointBit[p.col()];
        else                 mask[p.row()] |=  cPointBit[p.col()];
    }

public:
    
    // (1 = stone, 0 = empty)
    ULONG mask[ bSize1D ];
    
    void setm( const Point& p, SBYTE val )
    {
        if ( !p.ok() ) return;
        nc_set( p, val );
        mask_set( p, val );
    }
    
    StoneBoardMasked( const StoneBoard& board ) : StoneBoard( board )
    {
        for ( int p = 0; p < cBoardSize; p++ )
            mask_set( p, board.nc_get( p ) );
    }
};


// avdanced board, provides special accessors for influence calculation
class PowerBoard: public IBoard<short>
{
protected:
    virtual char get_symbol( int p, void* refData = NULL ) const
    { 
        if ( nc_get(p) > 0 ) return '+';
        if ( nc_get(p) < 0 ) return '-';
        return '.'; 
    }

public:
    
    void clear() { memset( data_, 0, sizeof(data_) ); }
    
    void inc( const Point& p, int val = 1 )
    {
        if ( p.ok() ) data_[p] += val;
    }

    void nc_inc( const Point& p, int val = 1 )
    { 
        data_[p] += val;
    }

    bool neitral( const Point& p ) const
    {
        return p.ok() && data_[p] > -25 && data_[p] < 25;
    }

    bool control( const Point& p, int color ) const
    {
        return p.ok() && (( data_[p] >  100 && color == eBlack ) || 
                          ( data_[p] < -100 && color == eWhite ));
    }

    int nc_color( int p ) const
    {
        if ( data_[p] > 0 ) return eBlack;
        if ( data_[p] < 0 ) return eWhite;
        return eEmpty;
    }
};


class SpPowerBoard: public IBoard<double>
{
protected:
    virtual char get_symbol( int p, void* refData = NULL ) const
    { 
        if ( nc_get(p) > 0 ) return '+';
        if ( nc_get(p) < 0 ) return '-';
        return '.';
    }

};


// general marker board, used to find regions intersections
class MarkerBoard: public IBoard<char>
{
public:
    enum tMarkerState { eClear, eMarked };

    //void mark( const Point& p )                 { set( p, eMarked ); }
    //void mark( const PointArray& points )       { setvec( points, eMarked ); }
    //void clear( const Point& p )                { set( p, eClear ); }
    //void clear( const PointArray& points )      { setvec( points, eClear ); }
    bool marked( const Point& p ) const         { return p.ok() ? nc_get(p) == eMarked : false; }

    void nc_mark( const Point& p )              { nc_set( p, eMarked ); }
    void nc_mark( const PointArray& points )    { nc_setvec( points, eMarked ); }
    void nc_clear( const Point& p )             { nc_set( p, eClear ); }
    void nc_clear( const PointArray& points )   { nc_setvec( points, eClear ); }
    bool nc_marked( const Point& p ) const      { return nc_get(p) == eMarked; }
};


// sometimes we don't need to intersect point regions, but their object indexes
class MarkerArray
{
protected:
    char *array_;

public:
    enum tMarkerState { eClear = 0, eMarked };

    MarkerArray( int size )
    { array_ = new char[size]; memset( array_, eClear, size ); }

    ~MarkerArray() { delete [] array_; }

    void mark( int index )
    {
        // not safe, must be sure, that index in bounds
        array_[index] = eMarked;
    }

    bool marked( int index ) const
    { 
        // not safe, must be sure, that index in bounds
        return array_[index] == eMarked;
    }
};


#endif // BOARDS_H_INCLUDED_