//************************************************************************
//   Copyright (C) 2004-2006 Maxim I. Guisov                            //
//                                                                      //
// Permission to use, copy, modify, distribute and sell this software   //
// and its documentation for any purpose is hereby granted without fee, //
// provided that the above copyright notice appear in all copies and    //
// that both that copyright notice and this permission notice appear    //
// in supporting documentation. Author makes no representations         //
// about the suitability of this software for any purpose.              //
// It is provided "as is" without express or implied warranty.          //
//                                                                      //
//************************************************************************

#ifndef RPROPNEURON_H_INCLUDED_
#define RPROPNEURON_H_INCLUDED_

#include "RPropUtils.h"


class TheNeuron
{
public:
    double *weights_;       // input weights
    double *deltaWs_;
    double *curGrad_;       // dE/dW
    double *newGrad_;
    
    double &output_;
    double &nDelta_;
    
    double * const inputPtr_;
    int    inputSize_;

    // ctr
    TheNeuron( int inputSize, double * const inputs, double &output, double &nDelta ) 
        : inputPtr_( inputs ), output_( output ), nDelta_( nDelta ), inputSize_( inputSize )
    {
        const int arrSize = inputSize + 1;
        const int memSize = sizeof( double ) * ( arrSize );
        
        weights_ = new double[ arrSize ]; memset( weights_, 0, memSize );
        curGrad_ = new double[ arrSize ]; memset( curGrad_, 0, memSize );
        newGrad_ = new double[ arrSize ]; memset( newGrad_, 0, memSize );
        deltaWs_ = new double[ arrSize ]; memset( deltaWs_, 0, memSize );
    }

    // dtr
    ~TheNeuron()
    { 
        delete [] weights_;
        delete [] deltaWs_;
        delete [] curGrad_;
        delete [] newGrad_;
    }

    void apply()
    {
        double sum = -weights_[0];
        for ( int i = inputSize_; i > 0; i-- )
            sum += weights_[i] * inputPtr_[i-1];
        output_ = activator( sum );
    }

    void backpropGrad( double sum )
    {
        nDelta_ = sum * output_ * ( 1 - output_ );
        newGrad_[0] -= nDelta_; // bias

        for ( int j = inputSize_; j > 0; j-- )
            newGrad_[j] += nDelta_ * inputPtr_[j-1];
    }

    
    void freeze_net()
    {
        for ( int i = inputSize_; i >= 0; i-- )
        {
            deltaWs_[i] = max( deltaWs_[i] * n_minus, min_delta );
            curGrad_[i] = 0;
            newGrad_[i] = 0;
        }            
    }
    
    // basic RPROP update variant
    void update_rprop()
    {
        for ( int i = inputSize_; i >= 0; i-- )
        {
            double direction = curGrad_[i] * newGrad_[i];

            // if same direction...
            if ( direction > 0 )
            {
                deltaWs_[i] = min( deltaWs_[i] * n_plus, max_delta );
                weights_[i] -= sign( newGrad_[i] ) * deltaWs_[i];
                curGrad_[i] = newGrad_[i];
            }

            // if reverse direction...
            else if ( direction < 0 )
            {
                // rollback previous weight change
                weights_[i] += sign( curGrad_[i] ) * deltaWs_[i];
                deltaWs_[i] = max( deltaWs_[i] * n_minus, min_delta );
                curGrad_[i] = 0;
            }

            // first training epoch otherwise...
            else
            {
                weights_[i] -= sign( newGrad_[i] ) * deltaWs_[i];
                curGrad_[i] = newGrad_[i];
            }
            
            newGrad_[i] = 0;
        }
    }

    
    // optimization called iRProp+
    void update_irprop( bool snerrIncreased )
    {
        for ( int i = inputSize_; i >= 0; i-- )
        {
            double direction = curGrad_[i] * newGrad_[i];

            // if same direction...
            if ( direction > 0 )
            {
                deltaWs_[i] = min( deltaWs_[i] * n_plus, max_delta );
                weights_[i] -= sign( newGrad_[i] ) * deltaWs_[i];
                curGrad_[i] = newGrad_[i];
            }

            // if reverse direction...
            else if ( direction < 0 )
            {
                // difference from basic RProp in this line
                if ( snerrIncreased )
                    // rollback prev weight change only if global error increased
                    weights_[i] += sign( curGrad_[i] ) * deltaWs_[i];

                deltaWs_[i] = max( deltaWs_[i] * n_minus, min_delta );
                curGrad_[i] = 0;
            }

            // first training epoch otherwise...
            else
            {
                weights_[i] -= sign( newGrad_[i] ) * deltaWs_[i];
                curGrad_[i] = newGrad_[i];
            }
            
            newGrad_[i] = 0;
        }
    }
    
    // advanced SARProp (Simulated Annealing)
    // take expTE = 2 ^ (-T * epoch), rmsErr = root(netErr/N)
    // param1 = sarprop_k1 * expTE
    // param2 = sarprop_k2 * rmsErr * rmsErr
    // param3 = sarprop_k3 * rmsErr * expTE
    void update_sarprop( double param1, double param2, double param3 )
    {
        for ( int i = inputSize_; i >= 0; i-- )
        {
            // SARPROP variation of gradient
            // newGrad_[i] -= param1 * weights_[i];
            //newGrad_[i] -= sign( newGrad_[i] ) * param1 * fabs( weights_[i] );
            
            double direction = curGrad_[i] * newGrad_[i];
            if ( direction > 0 )
            {
                deltaWs_[i] = min( deltaWs_[i] * n_plus, max_delta );
                weights_[i] -= sign( newGrad_[i] ) * deltaWs_[i];
                curGrad_[i] = newGrad_[i];
            }
            else if ( direction < 0 )
            {
                //weights_[i] += sign( curGrad_[i] ) * deltaWs_[i];
                if ( deltaWs_[i] < param2 )
                {
                    double addWs = param3 * ( rand() % 1000 ) / 1000.0;
                    deltaWs_[i] = deltaWs_[i] * n_minus + addWs;
                }
                else
                {
                    deltaWs_[i] *= n_minus;
                }
                deltaWs_[i] = max( deltaWs_[i], min_delta );
                curGrad_[i] = 0;
            }
            else
            {
                weights_[i] -= sign( newGrad_[i] ) * deltaWs_[i];
                curGrad_[i] = newGrad_[i];
            }
            
            newGrad_[i] = 0;
        }
    }


    
    // when curErr < prevErr, uses usual rprop update fn
    // otherwise goes back halfway on each weight
    void update_conv_rprop( bool snerrIncreased )
    {
        if ( snerrIncreased )
        {
            for ( int i = inputSize_; i >= 0; i-- )
            {
                deltaWs_[i] = max( deltaWs_[i] * n_minus, min_delta );
                weights_[i] += sign(curGrad_[i]) * deltaWs_[i];
                newGrad_[i] = 0;
            }
            return;
        }

        for ( int i = inputSize_; i >= 0; i-- )
        {
            double direction = curGrad_[i] * newGrad_[i];
            if ( direction > 0 )
            {
                deltaWs_[i] = min( deltaWs_[i] * n_plus, max_delta );
                weights_[i] -= sign( newGrad_[i] ) * deltaWs_[i];
                curGrad_[i] = newGrad_[i];
            }
            else if ( direction < 0 )
            {
                weights_[i] += sign( curGrad_[i] ) * deltaWs_[i];
                deltaWs_[i] = max( deltaWs_[i] * n_minus, min_delta );
                curGrad_[i] = 0;
            }
            else
            {
                weights_[i] -= sign( newGrad_[i] ) * deltaWs_[i];
                curGrad_[i] = newGrad_[i];
            }
            newGrad_[i] = 0;
        }
    }
};

#endif // RPROPNEURON_H_INCLUDED_