//************************************************************************
//   Copyright (C) 2004-2006 Maxim I. Guisov                            //
//                                                                      //
// Permission to use, copy, modify, distribute and sell this software   //
// and its documentation for any purpose is hereby granted without fee, //
// provided that the above copyright notice appear in all copies and    //
// that both that copyright notice and this permission notice appear    //
// in supporting documentation. Author makes no representations         //
// about the suitability of this software for any purpose.              //
// It is provided "as is" without express or implied warranty.          //
//                                                                      //
//************************************************************************

#ifndef RPROPOPTIMIZED_H_INCLUDED_
#define RPROPOPTIMIZED_H_INCLUDED_

#include "RPropLayer.h"


class TheNetwork
{
public:

    enum NetType { RPROP, iRPROP, SARP10, SARP15, SARP20, CONVRP } type_;

    int         inputSize_;
    double     *input_;
    
    int         layersNum_;
    TheLayer   *layers_;

    bool        batchGoingOn_;
    double      trainError_;
    double      trainErrorPrev_;
    double      testError_;
    double      testErrorMin_;
    std::string filename_;

    int         trainApplyRunsCount_;
    int         testApplyRunsCount_;
    int         epochCount_;
    int         epochCountAfterConv_;

    void fillInputVector( const double* vector )
    { 
        if ( vector )
            memcpy( input_, vector, inputSize_ * sizeof( double ) );
    }

    void fillOutputVector( double* vector ) const
    {
        if ( vector )
            memcpy( vector, layers_[ layersNum_ - 1 ].output_, layers_[ layersNum_ - 1 ].layerSize_ * sizeof( double ) );
    }

    void reset()
    {
        for ( int n = 0; n < layersNum_; n++ )
        {
            TheLayer &ref = layers_[n];
            for ( int j = ref.layerSize_ - 1; j >= 0; j-- )
            {
                TheNeuron &neuron = *( ref.neurons_[j] );
                for ( int i = ref.inputSize_; i >= 0; i-- )
                {
                    neuron.weights_[i] = ( rand() % 1001 ) / 1000.0 - .5;
                    neuron.deltaWs_[i] = start_delta; 
                    neuron.curGrad_[i] = 0.0;
                }
            }
        }
        trainError_ = 0.0;
        trainErrorPrev_ = -1.0;
        testErrorMin_ = 100;
        epochCount_ = 0;
        epochCountAfterConv_ = 0;
    }
    
    void resetDeltaWs()
    {
        for ( int n = 0; n < layersNum_; n++ )
        {
            TheLayer &ref = layers_[n];
            for ( int j = ref.layerSize_ - 1; j >= 0; j-- )
            {
                TheNeuron &neuron = *( ref.neurons_[j] );
                for ( int i = ref.inputSize_; i >= 0; i-- )
                    neuron.deltaWs_[i] += ( ( rand() % 1001 ) / 1000.0 - .5 ) * start_delta; 
            }
        }
    }

    bool loadNetwork()
    {
		FILE* file = 0;
        double value = 0;

	    fopen_s( &file, filename_.c_str(), "rb" );

	    if ( !file ) 
            return false;

        try
        {
            if ( !fread( &value, sizeof( double ), 1, file ) || value != layersNum_ )
                throw "wrong layers number";
            
            if ( !fread( &value, sizeof( double ), 1, file ) || value != inputSize_ )
                throw "wrong input size";

            int n;
            for ( n = 0; n < layersNum_; n++ )
                if ( !fread( &value, sizeof( double ), 1, file ) || value != layers_[n].layerSize_ )
                    throw "wrong layer size";

            for ( n = 0; n < layersNum_; n++ )
                for ( int i = 0; i < layers_[n].layerSize_; i++ )
                {
                    TheNeuron& neuron = *( layers_[n].neurons_[i] );
                    for ( int j = 0; j <= neuron.inputSize_; j++ )
                        if ( !fread( &neuron.weights_[j], sizeof( double ), 1, file ) ||
                             !fread( &neuron.deltaWs_[j], sizeof( double ), 1, file ) ||
                             !fread( &neuron.curGrad_[j], sizeof( double ), 1, file ) )
                        {
                            throw "failed to retrieve data";
                        }
                }

            if ( !fread( &value, sizeof( double ), 1, file ) )
                throw "failed to retrieve trains count";
                
            epochCount_ = ( int ) value;

            if ( !fread( &trainErrorPrev_, sizeof( double ), 1, file ) )
                throw "failed to retrieve prev network error";

            trainError_ = trainErrorPrev_;
            testErrorMin_ = 100;
            epochCountAfterConv_ = 0;
        }
        catch ( const char *ex )
        {
            printf( "Error loading network: %s\n", ex );
            fclose( file );
            return false;
        }

        fclose( file );
        return true;
    }

    void loadAs( const std::string& filename, NetType type )
    {
        filename_ = filename;
        type_ = type;
        
        if ( !loadNetwork() )
            reset();
    }

    void saveNetwork()
    {
        FILE* file;
		fopen_s( &file, filename_.c_str(), "wb" );

        if ( !file )
            return;

        double value = ( double ) layersNum_;
        fwrite( &value, sizeof( double ), 1, file );

        value = ( double ) inputSize_;
        fwrite( &value, sizeof( double ), 1, file );

        int n;
        for ( n = 0; n < layersNum_; n++ )
        {
            value = ( double ) layers_[n].layerSize_;
            fwrite( &value, sizeof( double ), 1, file );
        }

        for ( n = 0; n < layersNum_; n++ )
            for ( int i = 0; i < layers_[n].layerSize_; i++ )
            {
                TheNeuron& neuron = *( layers_[n].neurons_[i] );
                for ( int j = 0; j <= neuron.inputSize_; j++ )
                {
                    fwrite( &neuron.weights_[j], sizeof( double ), 1, file );
                    fwrite( &neuron.deltaWs_[j], sizeof( double ), 1, file );
                    fwrite( &neuron.curGrad_[j], sizeof( double ), 1, file );
                }
            }

        value = ( double ) epochCount_;
        fwrite( &value, sizeof( double ), 1, file );
        fwrite( &trainError_, sizeof( double ), 1, file );

        fclose( file );
    }

    void saveNetworkOutput()
    {
        // output train and test error
        char buf[256];
        sprintf_s( buf, 256, "%s.out.scv", filename_.c_str() );
        
        FILE* file;

        if ( epochCount_ == 1 )
            fopen_s( &file, buf, "wt" );
        else
            fopen_s( &file, buf, "at" );

        fprintf( file, "%7.5g, %7.5g\n", 
            trainError_ / trainApplyRunsCount_, 
            testError_ / testApplyRunsCount_ );
        
        fclose( file );
    }

    void saveAs( const std::string& filename )
    {
        filename_ = filename;
        saveNetwork();
    }

    TheNetwork( const std::vector<int> &topology, const std::string& filename, NetType type )
        : batchGoingOn_( false ), trainError_( 0.0 ),
          trainErrorPrev_( -1.0 ), trainApplyRunsCount_( 0 ),
          filename_( filename ), type_( type )
    {
        if ( topology.size() < 2 )
            return;

        inputSize_ = topology[0];
        layersNum_ = topology.size() - 1;

        input_ = new double[ inputSize_ ];
        layers_ = new TheLayer[ layersNum_ ];
        layers_[0].allocate( topology[1], inputSize_, input_ );
        for ( int n = 1; n < layersNum_; n++ )
            layers_[n].allocate( topology[n+1], topology[n], layers_[n-1].output_ );

        if ( !loadNetwork() )
            reset();
    }

    ~TheNetwork()
    { 
        if ( input_ ) { delete [] input_; input_ = 0; }    
        if ( layers_ ) { delete [] layers_; layers_ = 0; }    
    }

    void apply( const double *desired, bool testRun )
    {
        for ( int n = 0; n < layersNum_; n++ )
            layers_[n].apply();
        
        if ( batchGoingOn_ && desired )
        {
            int n = layersNum_ - 1;
            for ( int i = layers_[n].layerSize_; i-- > 0; )
            {
                TheNeuron &neuron = *( layers_[n].neurons_[i] );
                double err = neuron.output_ - desired[i];
                if ( testRun )
                    testError_ += err * err;
                else
                {
                    trainError_ += err * err;
                    neuron.backpropGrad( err );
                }
            }

            if ( testRun )
            {
                testApplyRunsCount_++;
                return;
            }
            
            // starting with n = layersNum - 2;
            for ( --n; n >= 0; n-- )
            {
                double *sumNiPtr = layers_[n].sum_ni_;
                memset( sumNiPtr, 0, sizeof( double ) * layers_[n].layerSize_ );
                int i, imax = layers_[n].layerSize_ - 1;

                for ( int k = layers_[n+1].layerSize_ - 1; k >= 0; k-- )
                {
                    TheNeuron &neuron = *( layers_[n+1].neurons_[k] );
                    for ( i = imax; i >= 0; i-- )
                        sumNiPtr[i] += neuron.nDelta_ * neuron.weights_[i+1];
                }

                for ( i = imax; i >= 0; i-- )
                    layers_[n].neurons_[i]->backpropGrad( sumNiPtr[i] );
            }

            trainApplyRunsCount_++;
        }
    }

    void apply( const double* input, double* output, double* desired = 0, bool testRun = true )
    {
		if ( !input ) return;
		fillInputVector( input );
		apply( desired, testRun );
		if ( output ) fillOutputVector( output );
    }

    void update_rprop()
    {
        for ( int n = layersNum_; n-- > 0; )
            for ( int i = layers_[n].layerSize_; i-- > 0; )
                layers_[n].neurons_[i]->update_rprop();
    }

    void update_irprop()
    {
        bool errIncreased = ( trainError_ > trainErrorPrev_ && trainErrorPrev_ != -1 );
        for ( int n = layersNum_; n-- > 0; )
            for ( int i = layers_[n].layerSize_; i-- > 0; )
                layers_[n].neurons_[i]->update_irprop( errIncreased );
    }

    void update_sarprop( double T )
    {
        //if ( trainError_ < trainErrorPrev_ * 0.95 )
        //{
        //    for ( int n = layersNum_; n-- > 0; )
        //        for ( int i = layers_[n].layerSize_; i-- > 0; )
        //            layers_[n].neurons_[i]->freeze_net();
        //    return;
        //}

        double expTE = exp( -T * log( 2.0 ) * epochCount_ );
        double rmsErr = sqrt( trainError_ / trainApplyRunsCount_ );
        double p1 = sarprop_k1 * expTE;
        double p2 = sarprop_k2 * rmsErr * rmsErr;
        double p3 = sarprop_k3 * rmsErr * expTE;
        //double saTerm = expTE * rmsErr;
        //double p1 = sarprop_k1 * saTerm;
        //double p2 = sarprop_k2 * saTerm * saTerm;
        //double p3 = sarprop_k3 * saTerm * saTerm;

        //printf("p1 = %.5g, p2 = %.5g, p3 = %.5g\n", p1, p2, p3);
        
        for ( int n = layersNum_; n-- > 0; )
            for ( int i = layers_[n].layerSize_; i-- > 0; )
                layers_[n].neurons_[i]->update_sarprop( p1, p2, p3 );
    }

    void update_conv_rprop()
    {
        bool errIncreased = ( trainError_ > trainErrorPrev_ && trainErrorPrev_ != -1 );

        for ( int n = layersNum_; n-- > 0; )
            for ( int i = layers_[n].layerSize_; i-- > 0; )
                layers_[n].neurons_[i]->update_conv_rprop( errIncreased );
    }

    void updateWeights()
    {
        switch ( type_ )
        {
            case RPROP: update_rprop(); break;
            case iRPROP: update_irprop(); break;
            case SARP10: update_sarprop( 0.010 ); break;
            case SARP15: update_sarprop( 0.015 ); break;
            case SARP20: update_sarprop( 0.020 ); break;
            case CONVRP: update_conv_rprop(); break;
            default:
                assert(0);
        }
    }

    void startBatchTraining()
    {
        batchGoingOn_ = true;
        trainError_ = 0.0;
        testError_ = 0.0;
        trainApplyRunsCount_ = 0;
        testApplyRunsCount_ = 0;
    }

    void finishBatchTraining()
    {
        batchGoingOn_ = false;
        if ( trainApplyRunsCount_ )
        {
            epochCount_++;
            updateWeights();
            if ( ( epochCount_ == 1 ) || ( testError_ < testErrorMin_ ) )
            {
                testErrorMin_ = testError_;
                saveNetwork();
                epochCountAfterConv_ = 0;
            }
            else
            {
                epochCountAfterConv_++;
            }
            
            printf(" %d ", epochCountAfterConv_);
            //saveNetworkOutput();
            trainErrorPrev_ = trainError_;
        }
    }

    double netTrainError() const { return trainApplyRunsCount_ ? trainError_ / trainApplyRunsCount_ : 0.0; }
    double netTestError() const { return testApplyRunsCount_ ? testError_ / testApplyRunsCount_ : 0.0; }
    int netEpochCount() const { return epochCount_; }
    bool netConverged() const { return epochCountAfterConv_ >= 100; }
};



#endif // RPROPOPTIMIZED_H_INCLUDED_

