//************************************************************************
//   Copyright (C) 2004-2006 Maxim I. Guisov                            //
//                                                                      //
// Permission to use, copy, modify, distribute and sell this software   //
// and its documentation for any purpose is hereby granted without fee, //
// provided that the above copyright notice appear in all copies and    //
// that both that copyright notice and this permission notice appear    //
// in supporting documentation. Author makes no representations         //
// about the suitability of this software for any purpose.              //
// It is provided "as is" without express or implied warranty.          //
//                                                                      //
//************************************************************************
#include "StdAfx.h"
#include "MorphEngine.h"
#include "..\\Navigator\\BoardNavigator.h"

// PURPOSE: counts all adjacent points on the board and fill neibours
void MorphEngine::GetNeibours( const PointArray& points, PointArray* neibours )
{
    assert( neibours );
    if ( !neibours ) return;
    
    ULONG bb[bSize1D], tmp[bSize1D];

    bb_clear(bb);
    bb_set_points(bb, points);  // bb  (( points
    bb_dilate(tmp, bb);         // tmp (( points + neibours
    bb_and_not(tmp, bb);        // tmp (( neibours
    bb_get_points(tmp, neibours);
}


// PURPOSE: given the block, counts its first, second and third order libersties
// and fills corresponding arrays; due to incremental update functionality,
// block->navi->board should always has filled mask, block->points should correspond
// to the board position and also, first order liberties should be presented
void MorphEngine::GetLiberties( Block* block )
{
    assert( block );
    if ( !block ) return;

    ULONG bb[bSize1D], libs[bSize1D], temp[bSize1D], mask[bSize1D];

    bb_clear(bb);
    bb_set_points(bb, block->points);       // bb   (( points
    bb_copy(mask, block->navi->board.mask); // mask (( stones on the board

    if ( block->liberties[0].size() ) {     // if we have first order liberties counted
        bb_set_points(libs, block->liberties[0]); // libs (( FOL
    } else {                                // otherwise...
        bb_dilate(libs, bb);                // libs (( points + neibours
        bb_and_not(libs, mask);             // libs (( empty neibours = first order liberties (FOL)
        bb_get_points(libs, &block->liberties[0]);
    }

    // count second order liberties
    bb_copy(temp, libs);            // temp (( FOL
    bb_or(bb, temp);                // bb   (( points + FOL
    bb_dilate(libs, bb);            // libs (( points + FOL + neibours
    bb_and_not(libs, mask);         // libs (( FOL + second order liberties (SOL)
    bb_and_not(libs, temp);         // libs (( SOL
    bb_get_points(libs, &block->liberties[1]);

    // count third order liberties
    bb_or(temp, libs);              // temp (( FOL + SOL
    bb_or(bb, temp);                // bb   (( points + FOL + SOL
    bb_dilate(libs, bb);            // libs (( points + FOL + SOL + neibours
    bb_and_not(libs, mask);         // libs (( FOL + SOL + third order liberties (TOL)
    bb_and_not(libs, temp);         // libs (( TOL
    bb_get_points(libs, &block->liberties[2]);
}


// PURPOSE: uses four dilation operations to calculate influence of the stones
void MorphEngine::EstimateInfluence( PowerBoard* power, const StoneBoard& board )
{
    // todo: make influence estimation depend on the blocks strength
    board_init( power, board );
    for (int d = 0; d < 4; d++) board_dilate_z( power );
}


// PURPOSE: uses dilation and erasion operators to estimate territory on the given board
void MorphEngine::EstimateTerritory( PowerBoard* terri, const StoneBoard& board )
{
    // todo: make territory estimation depend on the blocks strength
    board_init( terri, board );
    int d = 0, e = 0;
    for ( ; d < 4; d++ ) 
        board_dilate_z( terri );

    for ( ; e < d*(d-1)+1; e++ )
        board_erase_z( terri );
}


void MorphEngine::PropagatePower( SpPowerBoard* result, const StoneBoard& board )
{
    BoardNavigator navi( board, 0 );
    navi.Preprocess();
    PropagatePower( result, navi );
}

void MorphEngine::PropagatePower( SpPowerBoard* result, BoardNavigator& navi )
{
    // init board with present stones power
    for ( int bi = navi.sbIndex.size(); bi-- > 0; )
    {
        Block* block = navi.nc_get_block( navi.sbIndex[bi] );

        double power = block->predicted_status;
        if ( block->color() == eWhite )
            power *= -1.0;

        result->nc_setvec( block->points, power );
    }

    // empty points set to 0 power
    for ( int si = navi.esIndex.size(); si-- > 0; )
    {
        ESpace* esp = navi.nc_get_espace( navi.esIndex[si] );
        result->nc_setvec( esp->points, 0.0 );
    }

    for ( int iter = 0; iter < 170; iter++ )
    {
        SpPowerBoard temp;

        for ( int pos = 0; pos < cBoardSize; pos++ )
        {
            Point p( pos );
            double val = result->nc_get(p);

            for ( int n = 0; n < 4; n++ )
            {
                Point adj( p.adjacent(n) );
                if ( !adj.ok() ) {
                    val += result->nc_get(p);
                    continue;
                }

                //double av = result->nc_get( adj );
                //if ( av * val < 0 ) continue;

                val += result->nc_get( adj );
            }

            temp.nc_set( p, val );
        }

        result->copy( temp );
        result->print();
        printf("\n\n\n");
    }
}


void MorphEngine::bb_clear( ULONG src[bSize1D] )
{
    memset( src, 0, sizeof(ULONG) * bSize1D );
}


void MorphEngine::bb_set_points( ULONG bb[bSize1D], const PointArray& points )
{
    PointArray::const_iterator pi;
    for (pi = points.begin(); pi != points.end(); pi++)
        bb[ pi->row() ] |= cPointBit[ pi->col() ];
}


void MorphEngine::bb_get_points( ULONG const bb[bSize1D], PointArray* points )
{
    assert( points );
    points->clear();

    for (int row = 0, pos = 0; row < bSize1D; row++)
        for (int col = 0; col < bSize1D; col++, pos++) {
            if ( bb[row] & cPointBit[col] ) points->push_back(pos);
        }
}


void MorphEngine::bb_dilate( ULONG dst[bSize1D], ULONG const src[bSize1D] )
{
    bb_clear(dst);

    const ULONG *ps = src;
    ULONG *pd = dst, *pde = pd + bSize1D-1;

    if ( *ps ) *pd |= (*ps | (*ps >> 1) | (*ps << 1)) & cPointMask;
    *pd |= *(++ps);

    while (++pd != pde)
    {
        *pd |= *(ps-1);
        if ( *ps ) *pd |= (*ps | (*ps >> 1) | (*ps << 1)) & cPointMask;
        *pd |= *(++ps);
    }

    *pd |= *(ps-1);
    if ( *ps ) *pd |= (*ps | (*ps >> 1) | (*ps << 1)) & cPointMask;
}


void MorphEngine::bb_copy( ULONG dst[bSize1D], ULONG const src[bSize1D] )
{
    memcpy( dst, src, sizeof(ULONG) * bSize1D );
}


void MorphEngine::bb_and_not( ULONG op1[bSize1D], ULONG const op2[bSize1D] )
{
    for (register int i = bSize1D-1; i >= 0; i--)
        op1[i] &= !op2[i];
}


void MorphEngine::bb_or( ULONG op1[bSize1D], ULONG const op2[bSize1D] )
{
    for (register int i = bSize1D-1; i >= 0; i--)
        op1[i] |= op2[i];
}


void MorphEngine::board_init( PowerBoard* index, const StoneBoard& board )
{
    assert( index );
    for (int p = 0; p < bSize2D; p++)
    switch ( board.get(p) ) {
        case eEmpty: index->set( p,   0 ); break;
        case eBlack: index->set( p,  64 ); break;
        case eWhite: index->set( p, -64 ); break;
        default: assert(0);
    }
}


void MorphEngine::board_dilate_z( PowerBoard* index )
{
    assert( index );
    PowerBoard temp;

    for (int pos = 0; pos < bSize2D; pos++)
    {
        Point p(pos);
        int val = index->get(p);
        int black = 0, white = 0;

        for (int n = 0; n < 4; n++)
        {
            Point adj( p.adjacent(n) );
            if ( !adj.ok() ) continue;

            if ( index->get(adj) < 0 ) white++; else
            if ( index->get(adj) > 0 ) black++;
        }

        if ( black != 0 && white == 0 && val >= 0 ) 
            temp.set(p, val + black); 
        else if ( white != 0 && black == 0 && val <= 0 ) 
            temp.set(p, val - white); 
        else
            temp.set(p, val);
    }

    index->copy(temp);
}


void MorphEngine::board_erase_z( PowerBoard* index )
{
    assert( index );
    PowerBoard temp;

    for (int pos = 0; pos < bSize2D; pos++)
    {
        Point p(pos);
        int val = index->get(p);
        if ( val == 0 ) continue;
        
        int black = 0, white = 0, empty = 0;
        for (int n = 0; n < 4; n++)
        {
            Point adj( p.adjacent(n) );
            if ( !adj.ok() ) continue;

            if ( index->get(adj) < 0 ) white++; else
            if ( index->get(adj) > 0 ) black++; else empty++;
        }
        
        int nval;
		if ( val > 0 )
		{
			nval = val - white - empty;
			if ( nval < 0 ) val = 0;
			else val = nval;
            //temp.set( p, val - min(white + empty, val) ); 
		}
        else if ( val < 0 )
		{
			nval = val + black + empty;
			if ( nval > 0 ) val = 0;
			else val = nval;
            //temp.set( p, val + min(black + empty, -val) ); 
		}
    }

    index->copy(temp);
}
